/* ----------------------------------------------------------------------------
view_ = representing html grid
controller_ = for translating view to model
model_ = needed for 500811 command
---------------------------------------------------------------------------- */

//const VIDEO_WIDTH = 1280;
//const VIDEO_HEIGHT = 720;
const VIDEO_WIDTH = 1920;
const VIDEO_HEIGHT = 1080;

const CM_TO_INCH = 0.393701;

class MultiviewGrid
{
    constructor(rows, cols) 
    {
      this.viewRows = rows;
      this.viewCols = cols;
      this.viewDevices = [];
      this.viewTop;  // initial coords
      this.viewLeft;

      this.viewSizeX;
      this.viewSizeY;

      this.controllerGrid;

      this.modelTop = 0;  // initial coords for videowall/multiview in px,
      this.modelLeft = 0; // should always be (0,0)
      this.modelHeight;
      this.modelWidth;
    }

    getDevice(mac)
    {
        var device = null;
        for (key in this.viewDevices)
        {
            if (mac == this.viewDevices[key].mac) 
            {
                device = this.viewDevices[key];
                break;
            }
        }
        return device;
    }

    calculateWindows()
    {
        for (key in this.viewDevices)
        {
            if (this.viewDevices[key] instanceof MultiviewDisplay)
            {
                var modelDisplay = this.viewDevices[key];
                
                if (!isNaN(modelDisplay.viewTop) && !isNaN(modelDisplay.viewLeft))
                {                    
                    this.calculateDisplayWindows(modelDisplay.mac);
                }
            }
        }
    }

    checkOverlap()
    {
        for (key in this.viewDevices)
        {
            if (this.viewDevices[key] instanceof MultiviewDisplay)
            {
                var modelDisplay = this.viewDevices[key];

                if (!isNaN(modelDisplay.viewTop) && !isNaN(modelDisplay.viewLeft))
                {                    
                    this.calculateDisplayWindows(modelDisplay.mac);
                }
            }
        }
    }

    calculateDisplayWindows(macRX)
    {
        //var modelDisplay = this.viewDevices[macRX];
        var modelDisplay = this.getDevice(macRX);
        modelDisplay.removeAllWindows();

        for (key in this.viewDevices)
        {
            if (this.viewDevices[key] instanceof MultiviewSource)
            {
                var modelSource = this.viewDevices[key];

                if (!isNaN(modelSource.viewTop) && !isNaN(modelSource.viewLeft))
                {
                    var window = this.generateWindow(modelDisplay, modelSource);
                    if (window instanceof MultiviewWindow) modelDisplay.addWindow(window);
                }
            }
        }
    }

    generateWindow(display,source)
    {
        var debuglog = false;

        var l1x = ((display.viewLeft - this.viewLeft) / display.viewWidth) * VIDEO_WIDTH;
        var l1y = ((display.viewTop - this.viewTop) / display.viewHeight) * VIDEO_HEIGHT; 
        var r1x = ((display.viewLeft+(display.viewWidth) - this.viewLeft) / display.viewWidth) * VIDEO_WIDTH;
        var r1y = ((display.viewTop+(display.viewHeight) - this.viewTop) / display.viewHeight) * VIDEO_HEIGHT;

        var l2x = ((source.viewLeft - this.viewLeft) / display.viewWidth) * VIDEO_WIDTH;
        var l2y = ((source.viewTop - this.viewTop) / display.viewHeight) * VIDEO_HEIGHT;
        var r2x = ((source.viewLeft+(source.viewWidth) - this.viewLeft) / display.viewWidth) * VIDEO_WIDTH;
        var r2y = ((source.viewTop+(source.viewHeight) - this.viewTop) / display.viewHeight) * VIDEO_HEIGHT;

        // if source is not on display check
        if ((Math.max(0, Math.min(r1x,r2x) - Math.max(l1x,l2x)) * Math.max(0, Math.min(r1y,r2y) - Math.max(l1y,l2y)) === 0)) return false;
   
        // Compensations ******************************************************
        
        if ( (r2y-l2y) > VIDEO_HEIGHT )
        {
            if (debuglog) console.log("Source Height = " + (r2y-l2y));            

            if ( (r2y-l2y) == (VIDEO_HEIGHT*2) )
            {
                l1y = ((display.viewTop - this.viewTop) / display.viewHeight) * 1088; 
                r1y = ((display.viewTop+(display.viewHeight) - this.viewTop) / display.viewHeight) * 1088;
                l2y = ((source.viewTop - this.viewTop) / display.viewHeight) * 1088;
                r2y = ((source.viewTop+(source.viewHeight) - this.viewTop) / display.viewHeight) * 1088;

                if (debuglog) console.log("Compensating Height -8");
                l2y = l2y - 8;
                r2y = r2y - 8;
            }

            if ( (r2y-l2y) == (VIDEO_HEIGHT*4) )
            {            
                l1y = ((display.viewTop - this.viewTop) / display.viewHeight) * 1088; 
                r1y = ((display.viewTop+(display.viewHeight) - this.viewTop) / display.viewHeight) * 1088;
                l2y = ((source.viewTop - this.viewTop) / display.viewHeight) * 1088;
                r2y = ((source.viewTop+(source.viewHeight) - this.viewTop) / display.viewHeight) * 1088;

                if (debuglog) console.log("Compensating Height -4");
                l2y = l2y - 25;
                r2y = r2y - 25;
            }            
        }        

        /*
        console.log("Size:");
        console.log(Math.max(0, Math.min(r1y,r2y) - Math.max(l1y,l2y)));
        console.log("Y:");
        console.log(parseInt(Math.max(l1y,l2y) - l1y));
        console.log("+:");
        console.log((Math.max(0, Math.min(r1y,r2y) - Math.max(l1y,l2y))) + (Math.max(l1y,l2y) - l1y));
        console.log("source height:");
        console.log(r2y - l2y);
        */
        
        if ( (r2y - l2y) > (Math.max(0, Math.min(r1y,r2y) - Math.max(l1y,l2y)))) // source bigger than window
        {
            if ( Math.max(0, Math.min(r1y,r2y) - Math.max(l1y,l2y)) == (VIDEO_HEIGHT/2) ) // window is half size
            { 
                if ( (parseInt(Math.max(l1y,l2y) - l1y) == 0) )  //at the top            
                {
                    if (debuglog) console.log("Compensating Height +8");
                    l2y = l2y + 8;
                }
            }
        }

        if  ( ((Math.max(0, Math.min(r1y,r2y) - Math.max(l1y,l2y))) + (Math.max(l1y,l2y) - l1y)) == VIDEO_HEIGHT ) //at the bottom
        {
            if (debuglog) console.log("Compensating Height -2");          
            r2y = r2y - 2; 

        }
        // ********************************************************************

        //Shift/Scale calculations        
        var pixelShift = 8;
        l2x = l2x - (source.shiftLeft*pixelShift) + (source.shiftRight*pixelShift);
        r2x = r2x - (source.shiftLeft*pixelShift) + (source.shiftRight*pixelShift);
        l2y = l2y - (source.shiftUp*pixelShift) + (source.shiftDown*pixelShift);
        r2y = r2y - (source.shiftUp*pixelShift) + (source.shiftDown*pixelShift);   
        
        var pixelScale = 8;
        r2x = r2x - (source.scaleDownH*pixelScale) + (source.scaleUpH*pixelScale);
        r2y = r2y - (source.scaleDownV*pixelScale) + (source.scaleUpV*pixelScale); 
                   
        var pixelModding = 8;

        // ********************************************************************
        
        //if (debuglog) console.log("windowY % 8 = " + (parseInt(Math.max(l1y,l2y) - l1y) % 8));
        //if (debuglog) console.log("windowHeight % 8 = " + (parseInt(Math.max(0, Math.min(r1y,r2y) - Math.max(l1y,l2y))) % 8));
       
        // ********************************************************************        
        
        if (debuglog) console.log("Coords:")
        if (debuglog) console.log(l1x + " " + l1y + " " + r1x + " " + r1y);
        if (debuglog) console.log(l2x + " " + l2y + " " + r2x + " " + r2y);

        var windowViewWidth = Math.max(0, Math.min(r1x,r2x) - Math.max(l1x,l2x));
        var windowViewHeight = Math.max(0, Math.min(r1y,r2y) - Math.max(l1y,l2y));

        if ((windowViewWidth * windowViewHeight) === 0) return false;        

        var windowViewX = Math.max(l1x,l2x);
        var windowViewY = Math.max(l1y,l2y);

        var windowX = parseInt(windowViewX - l1x);
        var windowY = parseInt(windowViewY - l1y);

        var windowWidth = parseInt(windowViewWidth);
        var windowHeight = parseInt(windowViewHeight);

        if (debuglog) console.log( "windowY%8: " + (windowY % 8) );
        if (debuglog) console.log( "windowHeight%8: " + (windowHeight % 8) );

        var residue = (windowY % 8);
        if (residue >=0 && residue <= 3 )
        {
            if (debuglog) console.log("GOOD Y");
        }
        if (residue >=4 && residue <= 7 )
        {
            if (debuglog) console.log("BAD Y");
            windowY = Math.ceil(Math.ceil(windowY/pixelModding)*pixelModding);
        }

        residue = (windowHeight % 8);
        if (residue >=0 && residue <= 2 )
        {
            if (debuglog) console.log("GOOD HEIGHT");
        }
        if (residue >=3 && residue <= 7 )
        {
            if (debuglog) console.log("BAD HEIGHT");
            windowHeight = Math.ceil(Math.ceil(windowHeight/pixelModding)*pixelModding);            
        }

        /*
        windowX = parseInt(parseFloat(windowX/pixelModding)*pixelModding);
        windowY = Math.ceil(Math.ceil(windowY/pixelModding)*pixelModding);
        windowWidth = parseInt(parseFloat(windowWidth/pixelModding)*pixelModding);
        windowHeight = Math.ceil(Math.ceil(windowHeight/pixelModding)*pixelModding);
        */

        if (debuglog) console.log("DST:")
        if (debuglog) console.log(windowX + " " + windowY);
        if (debuglog) console.log(windowWidth + " " + windowHeight);        

        //var TempSourceWidth = ((source.viewWidth/display.viewWidth)*VIDEO_WIDTH) - (source.scaleDownH*pixelScale) + (source.scaleUpH*pixelScale); 
        //var TempSourceHeight = ((source.viewHeight/display.viewHeight)*VIDEO_HEIGHT) - (source.scaleDownV*pixelScale) + (source.scaleUpV*pixelScale);
        
        var TempSourceWidth  = (r2x-l2x); 
        var TempSourceHeight = (r2y-l2y);         
    
        //TempSourceWidth = parseInt(parseFloat(TempSourceWidth/pixelModding)*pixelModding);
        //TempSourceHeight = parseInt(parseFloat(TempSourceHeight/pixelModding)*pixelModding);

        var windowXRatio = parseFloat(((windowViewX-l2x)/TempSourceWidth));//.toPrecision(5);
        var windowYRatio = parseFloat(((windowViewY-l2y)/TempSourceHeight));//.toPrecision(5);

        var windowWidthRatio = ((windowViewWidth/TempSourceWidth));//.toPrecision(5);
        var windowHeightRatio = ((windowViewHeight/TempSourceHeight));//.toPrecision(5);

        if (debuglog) console.log("RATIO:")
        if (debuglog) console.log(windowXRatio + " " + windowYRatio);
        if (debuglog) console.log(windowWidthRatio + " " + windowHeightRatio); 
    
        // Pixel  calculations
            
        if (isNaN(display.screenSize) || (display.screenSize==0)) display.screenSize = 21;
        
        //Bezel calculations
        var cmDiagonal = parseFloat(10/CM_TO_INCH) * display.screenSize;
        var c = Math.sqrt ( (cmDiagonal*cmDiagonal) / (VIDEO_WIDTH*VIDEO_WIDTH + VIDEO_HEIGHT*VIDEO_HEIGHT) );        
        var mmPixel = Math.floor ( c * 1000 ) / 1000;        
        var inchPixel = (mmPixel * (CM_TO_INCH/10));
                
        var bezelRight  = ( windowXRatio+windowWidthRatio < 0.99 ) ? display.bezelRight : 0;
        var bezelBottom = ( windowYRatio+windowHeightRatio < 0.99 ) ? display.bezelBottom : 0;
        var bezelLeft   = ( (windowXRatio-windowWidthRatio) >= 0 ) ? display.bezelLeft :  0;
        var bezelTop    = ( (windowYRatio-windowHeightRatio) >= 0 ) ? display.bezelTop : 0;

        if (display.measureUnit == "cm")
        {
            bezelRight = bezelRight * CM_TO_INCH;
            bezelBottom = bezelBottom * CM_TO_INCH;
            bezelLeft = bezelLeft * CM_TO_INCH;
            bezelTop = bezelTop * CM_TO_INCH;
        }

        var newWidth = windowWidth;
        var newHeight = windowHeight;

        //bezel left and right
        windowWidthRatio = ( (windowWidthRatio*newWidth)-((windowWidthRatio*(bezelLeft/inchPixel)))-((windowWidthRatio*(bezelRight/inchPixel))) ) / newWidth;
        windowXRatio = ( (windowXRatio*newWidth)+((windowXRatio*(bezelLeft/inchPixel))) ) / newWidth;

        //bezel top and bottom
        windowHeightRatio = ( (windowHeightRatio*newHeight)-((windowHeightRatio*(bezelTop/inchPixel)))-((windowHeightRatio*(bezelBottom/inchPixel))) ) / newHeight;
        windowYRatio = ( (windowYRatio*newHeight)+((windowYRatio*(bezelTop/inchPixel))) ) / newHeight;    

        if (debuglog) console.log("Bezeled RATIO:")
        if (debuglog) console.log(windowXRatio + " " + windowYRatio);
        if (debuglog) console.log(windowWidthRatio + " " + windowHeightRatio);                         

        var window = new MultiviewWindow(windowX,windowY,windowHeight,windowWidth,windowXRatio,windowYRatio,windowHeightRatio,windowWidthRatio);

        window.num_aud_on = source.num_aud_on;
        window.zorder = source.zorder;
        window.aspectCvrs = source.aspectCvrs;
        window.repeatable = source.repeatable;
        window.low_delay = source.low_delay;
        window.src_onplaying = source.modelSourceURL;

        return window;
    }

    verifyControllerGrid()
    {
        var tempGrid=new Array(this.viewRows)
        for (var i=0; i<this.viewRows; i++) 
        {
            tempGrid[i]=new Array(this.viewCols); 
        }

        var device;
        for (key in this.viewDevices)
        {
            device = this.viewDevices[key];

            if (!isNaN(device.viewTop) && !isNaN(device.viewLeft))
            {
                if (device.type == 'display' || device.type == 'source')
                {
                    for (var i=device.viewTop; i<(device.viewTop+device.viewHeight); i++)
                    {
                        for (var j=device.viewLeft; j<(device.viewLeft+device.viewWidth); j++)
                        {
                            if ((i < this.viewRows) && (j < this.viewCols))
                            {                                
                                if (typeof tempGrid[i][j] == 'undefined')
                                {
                                    tempGrid[i][j] = new Array();
                                    tempGrid[i][j]['type'] = device.type;
                                }
                                else if ( device.type == 'source' && ( tempGrid[i][j]['type'] == 'source+source' || tempGrid[i][j]['type'] == 'both+source' ) )
                                {
                                    return "OVERLAPPING SOURCES";// source+source+source overlap
                                }                                   
                                else if (tempGrid[i][j]['type'] != device.type && tempGrid[i][j]['type'] != 'both' && tempGrid[i][j]['type'] != 'both+source') // you can overlap display/source, but not display/display or source/source
                                {
                                    tempGrid[i][j]['type'] = 'both';
                                }                                                                                
                                else if ( device.type == 'source' && ( tempGrid[i][j]['type'] == 'source' || tempGrid[i][j]['type'] == 'both' ) )  //ignore source ovelap
                                {
                                    tempGrid[i][j]['type'] = tempGrid[i][j]['type'].toString() + '+source';
                                }
                                else
                                {
                                    return "INCORRECT PLACEMENT"; //overlap
                                }                             
                            }
                            else
                            {
                                return "OUT OF BOUNDS"; //out of bounds
                            }                                
                        }
                    }
                }
            }    
        }

        //check size
        this.findInitialCoords();
        for (var keyRX in this.viewDevices)      
        {            
            if (this.viewDevices[keyRX] instanceof MultiviewDisplay)
            {                                
                var modelDisplay = this.viewDevices[keyRX];
                if (!isNaN(modelDisplay.viewTop) && !isNaN(modelDisplay.viewLeft))
                {
                    for (var keyTX in this.viewDevices)
                    {
                        if (this.viewDevices[keyTX] instanceof MultiviewSource)
                        {
                            var modelSource = this.viewDevices[keyTX];        
                            if (!isNaN(modelSource.viewTop) && !isNaN(modelSource.viewLeft))
                            {
                                var window = this.generateWindow(modelDisplay, modelSource);                              
                                if (window instanceof MultiviewWindow) 
                                {
                                    if ( (window.width <= 64) || (window.height <= 64) )
                                        return "SOURCE TOO SMALL"; //too small                                        
                                }
                            }
                        }
                    }
                }
            }
        }

        //console.log(tempGrid);
        return true;
        
    }

    addDevice(device)
    {        
        if (device instanceof MultiviewDevice)
        {
            //this.viewDevices[device.mac] = device;
            return this.viewDevices.push(device);            
        }
        else return false;
    }

    removeDevice(index)
    {
        this.viewDevices.splice(index,1);
    }

    removeAllDevices()
    {        
        this.viewDevices.splice(0,this.viewDevices.length); // or this.viewDevices = [];
        this.viewTop;
        this.viewLeft;
    }

    findInitialCoords()
    {
        var data = this.sortDevices('viewTop');        
        this.viewTop = parseInt(data[0].viewTop);
        var data = this.sortDevices('viewLeft');        
        this.viewLeft = parseInt(data[0].viewLeft);               
    }

    findSize()
    {
        this.viewSizeX = 0;
        this.viewSizeY = 0;

        this.findInitialCoords();
        for (key in this.viewDevices)
        {            
            if (!isNaN(this.viewDevices[key].viewTop) && !isNaN(this.viewDevices[key].viewLeft))
            {
                this.viewSizeX = Math.max( this.viewSizeX, ((this.viewDevices[key].viewLeft - this.viewLeft) + this.viewDevices[key].viewWidth) );
                this.viewSizeY = Math.max( this.viewSizeY, ((this.viewDevices[key].viewTop - this.viewTop) + this.viewDevices[key].viewHeight) );
            }
        }
    }

    sortDevices(parameter)
    {        
        var data = [];
        var i = 0;
        for (key in this.viewDevices)
        {
            data[i] = {};
            $.extend(true, data[i], this.viewDevices[key]);
            i++;
        }

        data.sort((function(index)
        {
            return function(a, b)
            {
                if (isNaN(a[index])) return 1;
                if (isNaN(b[index])) return -1;
                if (a[index] < b[index]) return -1;
                if (a[index] > b[index]) return 1;
                return 0;
            };
        })(parameter));

        return data;
    }

    printSize()
    {
        this.findSize();
        console.log('Size: (' + ((isNaN(this.viewSizeX)) ? "none" : this.viewSizeX.toString()) + ',' + ((isNaN(this.viewSizeY)) ? "none" : this.viewSizeY.toString()) + ')');
    }    

    printInitialCoords()
    {
        this.printInitialCoords();
        console.log('Initial Point: (' + ((isNaN(this.viewLeft)) ? "none" : this.viewLeft.toString()) + ',' + ((isNaN(this.viewTop)) ? "none" : this.viewTop.toString()) + ')');
    }    
}

class MultiviewWindow
{
    constructor( x=null, y=null, 
                 height=null, width=null, 
                 x_ratio="0.000000", y_ratio="0.000000", 
                 height_ratio="0.000000", width_ratio="0.000000" )
    {
        this.x; this.y;
        this.height; this.width;        

        this.x_ratio="0.000000"; this.y_ratio="0.000000";
        this.height_ratio="0.000000"; this.width_ratio="0.000000";

        this.src_onplaying = null;

        this.fill_attributes(x,y,height,width,x_ratio,y_ratio,height_ratio,width_ratio);
        
        this.num_aud_on = 0;
        this.zorder = 2;
        this.aspectCvrs = 0;
        this.low_delay = 0;
        this.rotate = 0;
        this.repeatable = 1;
    }

    getArray()
    {
        var data = [];
        
        //"window_zorder":"0","window_aspectCvrs":"1","window_rotate":"0"

        data['num_aud_on'] = parseInt(this.num_aud_on).toString();
        data['window_zorder'] = parseInt(this.zorder).toString();
        data['window_aspectCvrs'] = parseInt(this.aspectCvrs).toString();
        data['window_rotate'] = parseInt(this.rotate).toString();
        data['low_delay'] = parseInt(this.low_delay).toString();
        //data['repeatable'] = parseInt(this.repeatable).toString();            

        data['window_xcoord_dst'] = parseInt(this.x).toString();
        data['window_ycoord_dst'] = parseInt(this.y).toString();
        data['window_width_dst'] = parseInt(Math.ceil(this.width)).toString();
        data['window_height_dst'] = parseInt(Math.ceil(this.height)).toString();
        
        data['window_xcoord_ratio'] = parseFloat(this.x_ratio).toPrecision(5).toString();
        data['window_ycoord_ratio'] = parseFloat(this.y_ratio).toPrecision(5).toString();
        data['window_width_ratio'] = parseFloat(this.width_ratio).toPrecision(5).toString();
        data['window_height_ratio'] = parseFloat(this.height_ratio).toPrecision(5).toString();

        data['src_onplaying'] = this.src_onplaying.toString();
        //data['src_onplaying'] = '';

        return data;        
    }

    fill_attributes( x=this.x, y=this.y, height=this.height, width=this.width, x_ratio=this.x_ratio, y_ratio=this.y_ratio, height_ratio=this.height_ratio, width_ratio=this.width_ratio )
    {        

        this.x = Math.ceil(x).toString();
        this.y = Math.ceil(y).toString();
        this.height = Math.ceil(height).toString();
        this.width = Math.ceil(width).toString();
        this.x_ratio = parseFloat(x_ratio).toPrecision(5).toString();
        this.y_ratio = parseFloat(y_ratio).toPrecision(5).toString();
        this.height_ratio = parseFloat(height_ratio).toPrecision(5).toString();
        this.width_ratio = parseFloat(width_ratio).toPrecision(5).toString();
    }

    fill_source( url=null )
    {
        this.src_onplaying = url;
    }
}

class MultiviewDevice
{
    constructor(mac, type)
    {
        this.mac = mac;
        this.type = type;

        this.viewTop;
        this.viewLeft; 
        this.viewHeight = 1;
        this.viewWidth = 1;

        this.modelTop;
        this.modelLeft;
        this.modelHeight;
        this.modelWidth;
    }

    isInGrid()
    {
        if (!isNaN(this.viewTop) && !isNaN(this.viewLeft)) return true; else return false;
    }

    modifyCoord(x=null,y=null)
    {
        this.viewLeft = (x===null)? null : parseInt(x);
        this.viewTop = (y===null)? null : parseInt(y);        
    }

    printCoord()
    {
        console.log('Device Coords: (' + ((isNaN(this.viewLeft)) ? "false" : this.viewLeft.toString()) + ',' + ((isNaN(this.viewTop)) ? "false" : this.viewTop.toString()) + ')');
    }

    modifySize(width,height)
    {
        this.viewHeight = parseInt(height);
        this.viewWidth = parseInt(width);
        //this.printSize();
    }

    printSize()
    {
        console.log('Device Size: (' + ((isNaN(this.viewHeight)) ? "false" : this.viewHeight.toString()) + 'x' + ((isNaN(this.viewWidth)) ? "false" : this.viewWidth.toString()) + ')');
    }    

}

class MultiviewDisplay extends MultiviewDevice
{
    constructor(mac)
    {
        super(mac, 'display'); 

        this.modelWidth = VIDEO_WIDTH; //assume all displays as VIDEO_HEIGHTp
        this.modelHeight = VIDEO_HEIGHT;

        this.modelWindows = [];

        this.display_mode = 1;
        this.display_res = 100;
        this.deep_color = 0;
        this.num_aud_on = 0;

        this.bezelTop = 0.0;
        this.bezelBottom = 0.0;
        this.bezelLeft = 0.0;
        this.bezelRight = 0.0;

        this.screenSize = 21; // in inches
        this.screenWidth = 18.303;
        this.screenHeight = 10.2955;
        this.measureUnit = "inch";
    }

    getScreenSize()
    {
        this.screenSize =  Math.round( Math.sqrt( (this.screenWidth*this.screenWidth)+(this.screenHeight*this.screenHeight) ) );
        return (this.screenSize)
    } 
        

    setScreenSize(screenSize)
    {
        // 16x9 angle = 29.37
        // Cos(29.37) = 0.87    and   Sin(29.37) = 0.47 

        this.screenSize =  parseInt(screenSize);
        this.screenWidth = 0.87157 * parseFloat(this.screenSize);
        this.screenHeight = 0.49026 * parseFloat(this.screenSize);
    }

    getArray()
    {
        var data = [];

        data['macRx'] = this.mac.toString();

        data['display_mode'] = parseInt(this.display_mode).toString();
        data['display_res'] = parseInt(this.display_res).toString();
        data['deep_color'] = parseInt(this.deep_color).toString();
        data['num_aud_on'] = parseInt(this.num_aud_on).toString();

        data['mvCoordX'] = parseInt(this.viewLeft);
        data['mvCoordY'] = parseInt(this.viewTop);
        data['mvSizeX'] = parseInt(this.viewWidth);
        data['mvSizeY'] = parseInt(this.viewHeight);

        data['bezelTop'] = this.bezelTop;
        data['bezelBottom'] = this.bezelBottom;
        data['bezelLeft'] = this.bezelLeft;
        data['bezelRight'] = this.bezelRight;

        data['scaleUpH'] = 0;
        data['scaleDownH'] = 0;
        data['scaleUpV'] = 0;
        data['scaleDownV'] = 0;

        data['shiftUp'] = 0;
        data['shiftDown'] = 0;
        data['shiftRight'] = 0;
        data['shiftLeft'] = 0;        
                
        data['screenWidth'] = this.screenWidth;
        data['screenHeight'] = this.screenHeight;
        data['measureUnit'] = this.measureUnit;

        return data;        
    }

    hasWindows()
    {
        return !(this.modelWindows.length===0);
    }

    hasSource(url)
    {
        for (var key in this.modelWindows)
        {
            if (this.modelWindows[key].src_onplaying == url) return true;
        }
        return false;
    }

    removeAllWindows()
    {
        this.modelWindows.splice(0,this.modelWindows.length); // or this.modelWindows = [];
    }

    addWindow(x,y,height,width,x_ratio,y_ratio,height_ratio,width_ratio)
    {        
        this.modelWindows.push(new MultiviewWindow(x,y,height,width,x_ratio,y_ratio,height_ratio,width_ratio));
        return true;
    }

    addWindow(window)
    {
        if (window instanceof MultiviewWindow)
        { 
            this.modelWindows.push(window);
            return true;
        }
        else return false;
    }
}

class MultiviewSource extends MultiviewDevice
{    
    constructor(mac,url=null)
    {
        super(mac, 'source');

        this.zorder = 2;
        this.aspectCvrs = 0;
        this.rotate = 0;
        this.repeatable = 1;
        this.low_delay = 1;
        this.num_aud_on = 0;

        this.modelSourceURL = url;

        this.shiftUp = 0;
        this.shiftDown = 0;
        this.shiftRight = 0;
        this.shiftLeft = 0;

        this.scaleUpH = 0;
        this.scaleDownH = 0;
        this.scaleUpV = 0;
        this.scaleDownV = 0;
    }

    getArray()
    {
        var data = [];

        data['macTx'] = this.mac.toString();
        data['src_onplaying'] = this.modelSourceURL.toString();
        data['num_aud_on'] = this.num_aud_on;
        data['zorder'] = this.zorder;
        data['aspectCvrs'] = this.aspectCvrs;
        data['rotate'] = this.rotate;
        data['low_delay'] = this.low_delay;
        //data['repeatable'] = this.repeatable;

        data['mvCoordX'] = parseInt(this.viewLeft);
        data['mvCoordY'] = parseInt(this.viewTop);
        data['mvSizeX'] = parseInt(this.viewWidth);
        data['mvSizeY'] = parseInt(this.viewHeight);

        data['scaleUpH'] = parseInt(this.scaleUpH);
        data['scaleDownH'] = parseInt(this.scaleDownH);
        data['scaleUpV'] = parseInt(this.scaleUpV);
        data['scaleDownV'] = parseInt(this.scaleDownV);

        data['shiftUp'] = parseInt(this.shiftUp);
        data['shiftDown'] = parseInt(this.shiftDown);
        data['shiftRight'] = parseInt(this.shiftRight);
        data['shiftLeft'] = parseInt(this.shiftLeft);
        
        data['screenWidth'] = 0;
        data['screenHeight'] = 0;
        data['measureUnit'] = "inch";

        return data;        
    }


}