<?php
namespace AngelicaO;
use MNC\MNC_API;

class VideoWall {

    public function __construct() {
    }


    /**
     * Add a new video wall to the cached device list (meaning AngelicaO's own database)
     * @param $video_wall
     * @return array
     * @throws \Exception
     */
    public static function addVideoWall($video_wall) {
        if (empty($video_wall['customName'])) {
            return array(
                'status' => 'error',
                'message' => 'Name cannot be empty'
            );
        }
        if (!isset($video_wall['system_id'])) {
            return array(
                'status' => 'error',
                'message' => 'system_id cannot be empty'
            );
        }
        if (!isset($video_wall['videoWallFullConfigId'])) {
            return array(
                'status' => 'error',
                'message' => 'config identifier cannot be empty'
            );
        }
        if (empty($video_wall['videoWallSize'])) {
            return array(
                'status' => 'error',
                'message' => 'video wall size cannot be empty'
            );
        }
        $video_wall['mac'] = $video_wall['modelName'] = $video_wall['system_id'] . '_vw_' . $video_wall['videoWallFullConfigId'];

        $data = array(
            'customName' => $video_wall['customName'],
            'system_id' => $video_wall['system_id'],
            'mac' => $video_wall['mac'],
            'modelName' => $video_wall['modelName'],
            'isVideoWall' => true,
            'videoWallFullConfigId' => $video_wall['videoWallFullConfigId'],
            'videoWallSize' => implode('x', $video_wall['videoWallSize'])
        );

        //USE replace into instead of INSERT for cases like the 755 where the same product can be switched into RX or TX (and it's model name changes but not it's MAC)
        $sql = 'REPLACE INTO devices SET customName = ?:[customName,s], system_id = ?:[system_id,i], mac = ?:[mac,s], modelName = ?:[modelName,s], icon = "", location_id = NULL, deleted = 0, translate_x = NULL, translate_y = NULL, isVideoWall = true, videoWallFullConfigId = ?:[videoWallFullConfigId,s], videoWallSize = ?:[videoWallSize,s]';
        DB::query($sql, $data);

        if (!empty($video_wall['configs'])) {
            self::saveConfigsForVideoWall($video_wall);
        }
        self::saveDevicesForVideoWall($video_wall);

        return $data;
    }

    /**
     * Save the devices associated to each config, and to the main config
     * @param $video_wall
     * @throws \Exception
     */
    public static function saveDevicesForVideoWall($video_wall) {
        $devices = array();

        foreach ($video_wall['devices'] as $device) {
            $devices[] = array(
                'system_id' => $video_wall['system_id'],
                'config_id' => $video_wall['videoWallFullConfigId'],
                'mac' => $device['macRx'],
                'pos_x' => $device['xGlobalCoordonate'],
                'pos_y' => $device['yGlobalCoordonate']
            );
        }

        foreach ($video_wall['configs'] as $config) {
            foreach ($config['devices'] as $device) {
                $devices[] = array(
                    'system_id' => $video_wall['system_id'],
                    'config_id' => $config['vwConfigId'],
                    'mac' => $device['macRx'],
                    'pos_x' => $device['xGlobalCoordonate'],
                    'pos_y' => $device['yGlobalCoordonate']
                );
            }
        }

        //save the devices in each video wall config, including the main config
        foreach ($devices as $device) {
            $data = array(
                'system_id' => $device['system_id'],
                'config_id' => $device['config_id'],
                'mac' => $device['mac'],
                'pos_x' => $device['pos_x'],
                'pos_y' => $device['pos_y']
            );

            $sql = 'REPLACE INTO video_wall_config_devices SET system_id = ?:[system_id,i], config_id = ?:[config_id,i], mac = ?:[mac,s], pos_x = ?:[pos_x,i], pos_y = ?:[pos_y,i]';
            DB::query($sql, $data);
        }
    }

    /**
     * Saves the configs and devices in them for a video wall, identified by main config id
     * @param $video_wall
     * @internal param $configs
     * @throws \Exception
     */
    public static function saveConfigsForVideoWall($video_wall) {
        foreach ($video_wall['configs'] as $config) {
            $data = array(
                'system_id' => $video_wall['system_id'],
                'parent_config_id' => $video_wall['videoWallFullConfigId'],
                'config_id' => $config['vwConfigId'],
                'name' => $config['VideoWallName']
            );
            $sql = 'REPLACE INTO video_wall_configs SET system_id = ?:[system_id,i], parent_config_id = ?:[parent_config_id,i], config_id = ?:[config_id,i], name = ?:[name,s]';
            DB::query($sql, $data);
        }
    }

    public static function getVideoWalls($include_deleted = false) {
        $video_walls = array();

        //doesn't check user permissions as this is an internal function that needs to know all devices
        $sql = 'SELECT * FROM devices where isVideoWall = 1  ';
        if (empty($include_deleted)) {
            $sql .= ' and deleted != 1 ';
        }
        $_devices = DB::query($sql, array());

        foreach ($_devices as $device) {
            if (isset($device['videoWallFullConfigId']) && $device['videoWallFullConfigId'] !== null) {
                $video_wall = $device;

                $video_wall_configs = array();
                $data = array(
                    'system_id' => $device['system_id'],
                    'parent_config_id' => $device['videoWallFullConfigId']
                );
                $sql = 'SELECT * FROM video_wall_configs LEFT JOIN video_wall_config_devices ON video_wall_configs.config_id = video_wall_config_devices.config_id WHERE video_wall_configs.system_id = ?:[system_id,i] AND video_wall_configs.parent_config_id = ?:[parent_config_id,i]';
                $_video_wall_configs = DB::query($sql, $data);

                foreach ($_video_wall_configs as $config) {
                    if (!isset($video_wall_configs[$config['config_id']])) {
                        $video_wall_configs[$config['config_id']] = array(
                            'name' => $config['name'],
                            'devices' => array()
                        );
                    }

                    $video_wall_configs[$config['config_id']]['devices'][] =  array(
                        'mac' => $config['mac'],
                        'pos_x' => $config['pos_x'],
                        'pos_y' => $config['pos_y']
                    );
                }

                $video_wall['configs'] = $video_wall_configs;

                $video_walls[] = $video_wall;
            }
        }

        return $video_walls;
    }

    /**
     * Returns a video wall along with its subconfigs, and devices
     * @param null $mac
     * @return array|mixed
     * @throws \Exception
     */
    public static function getVideoWall($mac = null) {
        if (empty($mac)) {
            return array(
                'status' => 'error',
                'message' => 'You must provide a video wall identifier',
                'code' => 'VW.100'
            );
        }
        $video_wall = array();
        $user_account = MNC_API::getAccount();
        $data = array(
            'mac' => strtolower($mac),
            'user_id' => intval($user_account['user_id'])
        );


        $sql =
            'SELECT devices.*, video_wall_layouts.name as `connectedLayout`, video_wall_layouts.layout_id as `connectedLayoutId`, ' .
            '(SELECT COUNT(*) FROM video_wall_currently_playing_connections WHERE videoWallFullConfigId = devices.videoWallFullConfigId) as `connectedLayoutChanges` FROM devices ' .
            'LEFT JOIN video_wall_layouts ON (devices.selected_layout = video_wall_layouts.layout_id)';

        $sql .= ' WHERE LOWER(mac) = ?:[mac,s] AND isVideoWall != 0 AND mac NOT IN (select mac from blocked_devices WHERE user_id = ?:[user_id,i])';

        $_device = DB::query($sql, $data);
        if (!empty($_device)) {
            $_device = reset($_device);
            $video_wall = $_device;

            $video_wall['devices'] = self::getVideoWallDevices($_device['system_id'], $_device['videoWallFullConfigId']);

            $video_wall['configs'] = self::getVideoWallConfigs($_device['system_id'], $_device['videoWallFullConfigId']);
            //add in the full wall as a config by default
            $video_wall['configs'][$_device['videoWallFullConfigId']] = array(
                'config_id' => $_device['videoWallFullConfigId'],
                'name' => 'All TVs',
                'devices' => array()
            );
            foreach ($video_wall['devices'] as $device) {
                $video_wall['configs'][$_device['videoWallFullConfigId']]['devices'][] = array_merge($device, array('config_id' => $_device['videoWallFullConfigId']));
            }

            //end add in the full wall as a config by default

            $video_wall['layouts'] = self::getVideoWallLayouts($_device['system_id'], $_device['videoWallFullConfigId']);

            foreach ($video_wall['layouts'] as $layout_id => $layout_data) {
                $video_wall['layouts'][$layout_id]['layout_visual_grid'] = self::getLayoutVisualGrid($layout_id, $layout_data, $video_wall);
            }

            //get specific connections made on the video wall (even if a layout was selected, user can still make arbitrary connections)
            $video_wall['connectedLayoutSpecificConnections'] = self::getVideoWallSpecificConnections($video_wall['system_id'], $video_wall['videoWallFullConfigId']);
        }

        return $video_wall;
    }

    /**
     * Gets specific connections made on the video wall (even if a layout was selected, user can still make arbitrary connections)
     * @param $system_id
     * @param null $main_config_id
     * @return array
     * @throws \Exception
     */
    private static function getVideoWallSpecificConnections($system_id, $main_config_id = null) {
        if ($main_config_id === null || empty($system_id)) {
            return array(
                'status' => 'error',
                'message' => 'You must provide a video wall identifier',
                'code' => 'VW.100'
            );
        }
        $data = array(
            'system_id' => $system_id,
            'config_id' => $main_config_id
        );

        $sql = 'SELECT * FROM video_wall_currently_playing_connections WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[config_id,i]';

        $connections = DB::query($sql, $data);

        return $connections;

    }

    /**
     * Utility method that builds the layout's visual grid based on the video wall data, to be drawn by the UI
     * @param $layout_id
     * @param $layout_data
     * @param $video_wall_data
     * @return array
     */
    private static function getLayoutVisualGrid($layout_id, $layout_data, $video_wall_data) {
        $layout_visual_grid = array();
        $grid_size = explode('x', strtolower($video_wall_data['videoWallSize']));

        //initialize the visual grid to have the same amount of cells as the video wall
        for ($i = 0; $i < $grid_size[1]; $i++) {
            $cols = array();
            for ($j = 0; $j < sizeof($grid_size[0]); $j++) {
                $cols[] = array();
            }
            $layout_visual_grid[] = $cols;
        }


        //mark all the cells that are part of configs as such
        foreach ($layout_data['configs'] as $config_id) {
            if (!empty($video_wall_data['configs'][$config_id]['devices'])) {
                foreach ($video_wall_data['configs'][$config_id]['devices'] as $config_device) {
                    $layout_visual_grid[$config_device['pos_y']][$config_device['pos_x']] = array(
                        'isIndividualDisplay' => false,
                        'config_id' => $config_id,
                        'mac' => $config_device['mac']
                    );
                }
            }

        }

        //mark all the cells that are individual devices as such
        foreach ($video_wall_data['devices'] as $video_wall_device) {
            if (empty($layout_visual_grid[$video_wall_device['pos_y']][$video_wall_device['pos_x']])) {
                if ($layout_id === 'whole_layout') {
                    $layout_visual_grid[$video_wall_device['pos_y']][$video_wall_device['pos_x']] = array(
                        'isIndividualDisplay' => false,
                        'config_id' => $layout_id,
                        'mac' => $video_wall_device['mac']

                    );
                } else {
                    $layout_visual_grid[$video_wall_device['pos_y']][$video_wall_device['pos_x']] = array(
                        'isIndividualDisplay' => true,
                        'config_id' => false,
                        'mac' => $video_wall_device['mac']

                    );
                }
            }
            ksort($layout_visual_grid[$video_wall_device['pos_y']]);
        }


        if ($layout_id === 'whole_layout' && !empty($video_wall_data['bypass_layout_source'])) {//special case
            for ($row = 0; $row < sizeof($layout_visual_grid); $row++) {
                for ($col = 0; $col < sizeof($layout_visual_grid[$row]); $col++) {
                    $layout_visual_grid[$row][$col] = array(
                        'isIndividualDisplay' => false,
                        'config_id' => 'whole_layout',
                        'mac' => $video_wall_data['bypass_layout_source']
                    );
                }
            }
        }

        //now that each cell knows what config it belongs to, let's make it easier for the UI to know which "areas" are the same config. So that the icons can be placed nicely in the center of a rectangle area for example
        /**
         * sort cells by x, and then by y in sub array
         * for each cell which is a config cell (not individual TV)
         * if overflowed, ignore
         * if any cell with this config ID exists and is marked as overflowing, mark as don't show icon
         * loop through the row and determine the length of the rect on the X axis
         *   Loop through the row on the next col and determine if it's the same.
         *      if so, all overflowed. Mark current cell as overflowing, X,Y factor, z index higher than others
         *      else y factor is 1
         *
         *
         */
        for ($row = 0; $row < sizeof($layout_visual_grid); $row++) {
            for ($col = 0; $col < sizeof($layout_visual_grid[$row]); $col++) {
                if (!empty($layout_visual_grid[$row][$col]['isIndividualDisplay']) || emptynz($layout_visual_grid[$row][$col]['config_id'])) {
                    continue;
                }
                if (!empty($layout_visual_grid[$row][$col]['hideSourceIcon'])) {
                    continue;
                }

                $config_id = $layout_visual_grid[$row][$col]['config_id'];

                //loop through all other cells of this config and mark as overflowed since we only want to stretch the area of the first cell in a config
                for ($i = 0; $i < sizeof($layout_visual_grid); $i++) {
                    for ($y = 0; $y < sizeof($layout_visual_grid[$i]); $y++) {
                        if ($i != $row || $y != $col) {
                            if (!emptynz($layout_visual_grid[$i][$y]['config_id']) && $layout_visual_grid[$i][$y]['config_id'] == $config_id) {
                                $layout_visual_grid[$i][$y]['hideSourceIcon'] = true;
                            }
                        }
                    }
                }

                for ($col_end = $col + 1; $col_end < sizeof($layout_visual_grid[$row]); $col_end++) {
                    if ($layout_visual_grid[$row][$col_end]['config_id'] != $config_id) {
                        //$col_end--;
                        break;
                    }
                }
                $col_end--;

                //now we know at which x coordinate the area finishes, let's check the coming rows
                for ($row_end = $row + 1; $row_end < sizeof($layout_visual_grid); $row_end++) {
                    $rowMatches = true;//is the next row the same? Or are the TVs in this config arranged differently on the next row?
                    for ($next_row_col = $col; $next_row_col <= $col_end; $next_row_col++) {
                        if ($layout_visual_grid[$row_end][$next_row_col]['config_id'] != $config_id) {
                            $rowMatches = false;
                        }
                    }
                    if (empty($rowMatches)) {
                        break;
                    }
                }
                $row_end--;

                $layout_visual_grid[$row][$col]['overflowX'] = ($col_end - $col) + 1;
                $layout_visual_grid[$row][$col]['overflowY'] = ($row_end - $row) + 1;

            }
        }


        return $layout_visual_grid;
    }


    public static function getVideoWallDevices($system_id, $videoWallFullConfigId) {
        $devices = array();

        //get the devices for the video wall
        $data = array(
            'system_id' => $system_id,
            'config_id' => $videoWallFullConfigId
        );
        $sql = 'SELECT video_wall_config_devices.*, devices.customName FROM video_wall_config_devices LEFT JOIN devices ON LOWER(devices.mac) = LOWER(video_wall_config_devices.mac) WHERE video_wall_config_devices.system_id = ?:[system_id,i] AND video_wall_config_devices.config_id = ?:[config_id,i]';
        $_video_wall_devices = DB::query($sql, $data);
        $video_wall['devices'] = array();
        foreach ($_video_wall_devices as $device) {
            $devices[] = array(
                'mac' => $device['mac'],
                'pos_x' => $device['pos_x'],
                'pos_y' => $device['pos_y'],
                'customName' => $device['customName'],
            );
        }

        return $devices;
    }

    public static function getVideoWallLayouts($system_id, $videoWallFullConfigId) {
        //TODO check user permissions
        $layouts = array();

        //add in the default layout
        $layouts['full_config'] = array(//individual TVs
            'layout_id' => 'full_config',
            'configs' => array(),
            'name' => 'Individual TVs'
        );
        $layouts['whole_layout'] = array(//whole video wall
            'layout_id' => 'whole_layout',
            'configs' => array(),
            'name' => 'Full Video Wall'
        );

        //get layouts
        $data = array(
            'system_id' => $system_id,
            'videoWallFullConfigId' => $videoWallFullConfigId
        );
        $sql = 'SELECT * FROM video_wall_layouts LEFT JOIN video_wall_layout_configs USING(layout_id) WHERE video_wall_layouts.system_id = ?:[system_id,i] AND video_wall_layouts.videoWallFullConfigId = ?:[videoWallFullConfigId,i]';
        $_layouts = DB::query($sql, $data);
        foreach ($_layouts as $layout) {
            if (!isset($layouts[$layout['layout_id']])) {
                $layouts[$layout['layout_id']] = array(
                    'layout_id' => $layout['layout_id'],
                    'configs' => array(),
                    'name' => $layout['name'],
                    'source_selections' => array('configs' => array(), 'individualDisplays' => array())
                );

                //get the sources attached to this layout
                $sources_sql = 'SELECT * FROM video_wall_layout_sources WHERE layout_id = ?:[layout_id,i]';
                $_sources = DB::query($sources_sql, array('layout_id' => $layout['layout_id']));

                foreach ($_sources as $source) {
                    if ($source['object_type'] === 'config') {
                        $layouts[$layout['layout_id']]['source_selections']['configs'][$source['object_id']] = $source['source_mac'];
                    } else if ($source['object_type'] === 'display') {
                        $layouts[$layout['layout_id']]['source_selections']['individualDisplays'][$source['object_id']] = $source['source_mac'];
                    }
                }
            }

            $layouts[$layout['layout_id']]['configs'][] = $layout['config_id'];
        }

        return $layouts;
    }

    public static function getVideoWallConfigs($system_id, $videoWallFullConfigId) {
        //get the subconfigs with their devices
        $video_wall_configs = array();
        $data = array(
            'system_id' => $system_id,
            'parent_config_id' => $videoWallFullConfigId
        );
        $sql = 'SELECT video_wall_configs.*, video_wall_config_devices.*, devices.customName FROM video_wall_configs LEFT JOIN video_wall_config_devices ON video_wall_configs.config_id = video_wall_config_devices.config_id LEFT JOIN devices ON devices.mac = video_wall_config_devices.mac WHERE video_wall_configs.system_id = ?:[system_id,i] AND video_wall_configs.parent_config_id = ?:[parent_config_id,i]';
        $_video_wall_configs = DB::query($sql, $data);

        foreach ($_video_wall_configs as $config) {
            if (!isset($video_wall_configs[$config['config_id']])) {
                $video_wall_configs[$config['config_id']] = array(
                    'config_id' => $config['config_id'],
                    'name' => $config['name'],
                    'devices' => array()
                );
            }

            $video_wall_configs[$config['config_id']]['devices'][] = array(
                'mac' => $config['mac'],
                'pos_x' => $config['pos_x'],
                'pos_y' => $config['pos_y'],
                'customName' => $config['customName'],
                'config_id' => $config['config_id']
            );
        }

        return $video_wall_configs;
    }


    public static function updateVideoWall($fields) {

        if (!isset($fields['system_id']) || !isset($fields['videoWallFullConfigId'])) {
            return array(
                'status' => 'error',
                'message' => 'System or main config id cannot be empty'
            );
        }

        $data = array(
            'system_id' => $fields['system_id'],
            'videoWallFullConfigId' => $fields['videoWallFullConfigId'],
        );

        $sql = 'UPDATE devices SET ';
        $where_clauses = array();

        if (isset($fields['customName'])) {
            $where_clauses[] = ' customName = ?:[customName,s] ';
            $data['customName'] = $fields['customName'];
        }

        if (isset($fields['videoWallSize'])) {
            $where_clauses[] = ' videoWallSize = ?:[videoWallSize,s] ';
            $data['videoWallSize'] = implode('x', $fields['videoWallSize']);
        }

        $sql .= ' ' . implode(', ', $where_clauses) . ' WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[videoWallFullConfigId,i] LIMIT 1 ';
        DB::query($sql, $data);

        return array(
            'status' => 'success'
        );


    }


    /**
     * Save a layout for a video wall
     * @param $layout
     * @return array success or failure indicator
     * @throws \Exception
     */
    public static function saveVideoWallLayout($layout) {
        if (empty($layout) || empty($layout['videoWallFullConfigId']) || empty($layout['system_id']) || empty($layout['layout_name'])) {
            return array(
                'status' => 'error',
                'message' => 'incomplete layout information'
            );
        }

        if (empty($layout['layout_id'])) {
            $data = array(
                'system_id' => $layout['system_id'],
                'videoWallFullConfigId' => $layout['videoWallFullConfigId'],
                'name' => $layout['layout_name']
            );
            $sql = 'INSERT INTO video_wall_layouts SET system_id = ?:[system_id,i], videoWallFullConfigId = ?:[videoWallFullConfigId,i], name = ?:[name,s]';
            $layout_id = DB::query($sql, $data, array('return_insert_id' => true));
            $layout['layout_id'] = $layout_id;
        } else {
            $data = array(
                'layout_id' => $layout['layout_id'],
                'name' => $layout['layout_name']
            );
            $sql = 'UPDATE video_wall_layouts SET name = ?:[name,s] WHERE layout_id = ?:[layout_id,i]';
            DB::query($sql, $data);
        }


        //selected config data
        //clear the existing data first
        $data = array(
            'layout_id' => $layout['layout_id']
        );
        $sql = 'DELETE FROM video_wall_layout_configs WHERE layout_id = ?:[layout_id,i]';
        DB::query($sql, $data);
        if (!empty($layout['selected_configs'])) {
            //insert the new data
            foreach ($layout['selected_configs'] as $config) {
                $data = array(
                    'layout_id' => $layout['layout_id'],
                    'config_id' => $config['config_id']
                );
                $sql = 'INSERT INTO video_wall_layout_configs SET layout_id = ?:[layout_id,i], config_id = ?:[config_id,i]';
                DB::query($sql, $data);
            }
        }


        //source selection data
        //first delete old data
        $data = array(
            'layout_id' => $layout['layout_id']
        );
        $sql = 'DELETE FROM video_wall_layout_sources WHERE layout_id = ?:[layout_id,i]';
        DB::query($sql, $data);

        //now insert new data
        if (!empty($layout['source_selections'])) {
            foreach ($layout['source_selections']['configs'] as $config_id => $source_mac) {
                $data = array(
                    'layout_id' => $layout['layout_id'],
                    'object_id' => $config_id,
                    'object_type' => 'config',
                    'source_mac' => (empty($source_mac) ? '' : $source_mac)
                );

                $sql = 'INSERT INTO video_wall_layout_sources SET layout_id = ?:[layout_id,i], object_id = ?:[object_id,s], object_type = ?:[object_type,s], source_mac = ?:[source_mac,s]';
                DB::query($sql, $data);
            }

            foreach ($layout['source_selections']['individualDisplays'] as $display_mac => $source_mac) {
                $data = array(
                    'layout_id' => $layout['layout_id'],
                    'object_id' => $display_mac,
                    'object_type' => 'display',
                    'source_mac' => $source_mac
                );

                $sql = 'INSERT INTO video_wall_layout_sources SET layout_id = ?:[layout_id,i], object_id = ?:[object_id,s], object_type = ?:[object_type,s], source_mac = ?:[source_mac,s]';
                DB::query($sql, $data);
            }
        }


        return array(
            'status' => 'success'
        );

    }

    /**
     * Save a layout for a video wall
     * @param $system_id
     * @param $videoWallFullConfigId
     * @param $layout_id Either the ID (FK from layouts table) or the name of the auto-generated layouts like "whole_layout"
     * @return array success or failure indicator
     * @throws \Exception
     */
    public static function setSelectedLayout($system_id, $videoWallFullConfigId, $layout_id) {
        if (empty($layout_id) || empty($videoWallFullConfigId) || empty($system_id)) {
            return array(
                'status' => 'error',
                'message' => 'you must specify a layout and a video wall'
            );
        }

        $data = array(
            'layout_id' => $layout_id,
            'system_id' => $system_id,
            'videoWallFullConfigId' => $videoWallFullConfigId
        );
        $sql = 'UPDATE devices SET selected_layout = ?:[layout_id,s], bypass_layout_source = "" WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[videoWallFullConfigId,i]';
        DB::query($sql, $data);

        //clear the specific connections if any. Meaning, the connections that were made on top of the layout selection
        DB::query('DELETE FROM video_wall_currently_playing_connections WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[videoWallFullConfigId,i]', $data);

        return array(
            'status' => 'success'
        );

    }

    /**
     * Set specific connections for a video wall. This function merely saves which connections have already been done. It
     * doesn't actually make any connections. It is used to show the user what is currently playing in the video wall.
     *
     * @param $system_id
     * @param $videoWallFullConfigId
     * @param $connections array of connections
     * @return array success or failure indicator
     * @throws \Exception
     */
    public static function setVideoWallSpecificConnections($system_id, $videoWallFullConfigId, $connections) {
        if (empty($videoWallFullConfigId) || empty($connections) || empty($system_id)) {
            return array(
                'status' => 'error',
                'message' => 'you must specify connections and a video wall'
            );
        }

        foreach ($connections as $connection) {
            if (!empty($connection['config_id']) && $connection['config_id'] === 'whole_layout') {
                self::setFullVideoWallSource($system_id, $videoWallFullConfigId, $connection['source']);
                continue;
            }

            $data = array(
                'source' => $connection['source'],
                'system_id' => $system_id,
                'videoWallFullConfigId' => $videoWallFullConfigId
            );

            if (!empty($connection['config_id'])) {
                $data['object_type'] = 'config';
                $data['object_id'] = $connection['config_id'];
            } else if (!empty($connection['mac'])) {
                $data['object_type'] = 'tv';
                $data['object_id'] = $connection['mac'];
            } else {
                continue;
            }

            $sql = 'REPLACE INTO video_wall_currently_playing_connections SET source = ?:[source,s], system_id = ?:[system_id,i], videoWallFullConfigId = ?:[videoWallFullConfigId,i], object_type = ?:[object_type,s], object_id = ?:[object_id,s]';
            DB::query($sql, $data);
        }

        return array(
            'status' => 'success'
        );

    }

    /**
     * Deletes a layout and optionally clears the connectedLayout data
     * @param $layout_id
     * @param $options [bool clearConnectedLayoutData, int videoWallFullConfigId]
     * @return array
     * @throws \Exception
     */
    public static function deleteLayout($layout_id, $options) {
        if (empty($layout_id) && $layout_id !== 0) {
            return array(
                'status' => 'error',
                'message' => 'you must specify a layout to delete'
            );
        }

        $data = array(
            'layout_id' => $layout_id
        );
        $sql = 'DELETE FROM video_wall_layouts WHERE layout_id = ?:[layout_id,i]';
        DB::query($sql, $data);

        $sql = 'DELETE FROM video_wall_layout_sources WHERE layout_id = ?:[layout_id,i]';
        DB::query($sql, $data);

        $sql = 'DELETE FROM video_wall_layout_configs WHERE layout_id = ?:[layout_id,i]';
        DB::query($sql, $data);

        if (!empty($options['clearConnectedLayoutData']) && !empty($options['system_id']) && !empty($options['videoWallFullConfigId'])) {
            $data = array(
                'system_id' => $options['system_id'],
                'videoWallFullConfigId' => $options['videoWallFullConfigId']
            );
            $sql = 'UPDATE devices SET selected_layout = "" WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[videoWallFullConfigId,i]';
            DB::query($sql, $data);

            $sql = 'DELETE FROM video_wall_currently_playing_connections WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[videoWallFullConfigId,i]';
            DB::query($sql, $data);
        }

        return array(
            'status' => 'success'
        );
    }


    public static function setFullVideoWallSource($system_id, $videoWallFullConfigId, $source_mac) {
        if (empty($videoWallFullConfigId) || empty($source_mac) || empty($system_id)) {
            return array(
                'status' => 'error',
                'message' => 'you must specify a source and a video wall'
            );
        }

        //delete the selected layout and set the current full VW source (bypass field)
        $data = array(
            'system_id' => $system_id,
            'videoWallFullConfigId' => $videoWallFullConfigId,
            'source_mac' => $source_mac
        );
        $sql = 'UPDATE devices SET selected_layout = "", bypass_layout_source = ?:[source_mac,s] WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[videoWallFullConfigId,i]';
        DB::query($sql, $data);

        //clear the specific connections if any. Meaning, the connections that were made on top of the layout selection
        DB::query('DELETE FROM video_wall_currently_playing_connections WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[videoWallFullConfigId,i]', $data);

        return array(
            'status' => 'success'
        );
    }

    public static function processVideoWallPayloadFromMNC($video_walls) {

        if (empty($video_walls)) {
            $video_walls = array();
        }

        $all_video_walls = self::getVideoWalls(true);

        //clear the video wall subconfig and devices data for reinsertion right after
        DB::query('TRUNCATE TABLE video_wall_config_devices', array());
        DB::query('TRUNCATE TABLE video_wall_configs', array());

        if (!empty($video_walls)) {
            foreach ($video_walls as $video_wall) {
                //check if the device exists by main config id (meaning the config id of the "main" config. i.e the config that has all TVs assigned
                $exists = false;
                for ($i = 0; $i < sizeof($all_video_walls); $i++) {
                    if (strtolower($video_wall['vwConfigId']) == strtolower($all_video_walls[$i]['videoWallFullConfigId']) &&
                        intval($video_wall['system_id']) === intval($all_video_walls[$i]['system_id'])) {
                        $exists = $i;
                    }
                }
                if ($exists === false) {
                    self::addVideoWall(array(
                        'customName' => $video_wall['VideoWallName'],
                        'system_id' => $video_wall['system_id'],
                        'videoWallFullConfigId' => $video_wall['vwConfigId'],
                        'videoWallSize' => $video_wall['vwSize'],
                        'configs' => (isset($video_wall['configs']) ? $video_wall['configs'] : array()),
                        'devices' => (isset($video_wall['devices']) ? $video_wall['devices'] : array())
                    ));

                } else {

                    self::updateVideoWall(array(
                        'customName' => $video_wall['VideoWallName'],
                        'system_id' => $video_wall['system_id'],
                        'videoWallFullConfigId' => $video_wall['vwConfigId'],
                        'videoWallSize' => $video_wall['vwSize'],
                    ));
                    //re-save the config data in case there's new info

                    $video_wall['videoWallFullConfigId'] = $video_wall['vwConfigId'];
                    if (!empty($video_wall['configs'])) {
                        self::saveConfigsForVideoWall($video_wall);
                    }
                    //and re-save the devices of the video wall itself in case the TVs were moved around / removed
                    //TODO test such a case to see if works
                    self::saveDevicesForVideoWall($video_wall);
                }

            }
        }

        //check if any existing devices are no longer found in the MNC data
        foreach ($all_video_walls as $angelicao_device) {
            $exists = false;
            $isDisconnected = false;
            foreach ($video_walls as $mnc_device) {
                if (intval($angelicao_device['videoWallFullConfigId']) === intval($mnc_device['vwConfigId']) &&
                    intval($angelicao_device['system_id']) === intval($mnc_device['system_id'])) {
                    $exists = true;

                    if (!empty($angelicao_device['isDisconnected'])) {
                        $isDisconnected = true;//will be used below to unmark it as disconnected if necessary
                    }
                }

            }

            if (empty($exists)) {
                $data = array(
                    'isDisconnected' => true,
                    'fullConfigId' => $angelicao_device['videoWallFullConfigId'],
                    'system_id' => $angelicao_device['system_id']
                );
                //$sql = 'UPDATE devices SET isDisconnected = ?:[isDisconnected,b] WHERE videoWallFullConfigId = ?:[fullConfigId,i] AND isVideoWall = 1';
                $sql = 'DELETE FROM devices WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[fullConfigId,i] AND isVideoWall = 1 LIMIT 1';
                DB::query($sql, $data);

            } else {//if the device is already in the data

                if ($isDisconnected === true) {//if the device was previously marked as disconnected, undo that.
                    $data = array(
                        'isDisconnected' => false,
                        'fullConfigId' => $angelicao_device['videoWallFullConfigId'],
                        'system_id' => $angelicao_device['system_id']
                    );
                    $sql = 'UPDATE devices SET isDisconnected = ?:[isDisconnected,b] WHERE system_id = ?:[system_id,i] AND videoWallFullConfigId = ?:[fullConfigId,i] AND isVideoWall = 1';
                    DB::query($sql, $data);

                }
            }
        }

    }


}